#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string>
#include <sys/types.h>
#include <fstream>
#include <unistd.h>
#include <iostream>
#include <sys/inotify.h>
#include <time.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <boost/bind.hpp>
#include <boost/shared_ptr.hpp>
#include <vector>
#include <fcntl.h>
#include "QueueFilter.h"
#include "Inotify.h"
#include "xmlParser.h"
#include "QueueRetry.h"
#include "FileSynchronize.h"
#include "Initialize.h"
#include "main.h"
#include "SocketInterface.h"
#include "HttpInterface.h"
#include "CdnRefreshInterface.h"

#define SCRIPT_EXECUTE_INTERVAL 3600 //每隔1个小时执行一次rsync_fail_log.sh

int FileSynchronize::sleep_group = 0; //用于统计当前处于sleep状态的线程数量

boost::condition_variable FileSynchronize::work_cond; //工作条件对象

boost::mutex FileSynchronize::worklock; //用于统计工作线程数量的锁定

int FileSynchronize::m_crontab = -1; //默认不开起crontab进行整体推送

std::string FileSynchronize::watch; //

std::string FileSynchronize::m_plugin;

std::vector<ptrRmtServer> FileSynchronize::rmtServers;

FileSynchronize::FileSynchronize(ptrInitialize init, ptrQFilter qf, ptrQRetry qr)
{
	m_init = init;
	m_qf = qf;
	m_qr = qr; //还没有用，但暂留
	XmlParse( init ); //解析xml脚本
	if (init->exec_flag & RSYNC_ONCE)
	{
		FileSynchronize::RsyncOnce( );
	}
	ptrInterface itf;
	if (!m_plugin.empty( ))
	{
		if (m_plugin == "refreshCDN")
		{
			ptrInterface temp( new CdnRefreshInterface( ) );
			itf = temp;
		}
		if (m_plugin == "socket")
		{
			ptrInterface temp( new SocketInterface( ) );
			itf = temp;
		}
		if (m_plugin == "http")
		{
			ptrInterface temp( new HttpInterface( ) );
			itf = temp;
		}
		if (itf.use_count( ) != 0)
		{
			itf->XmlParse( init->config_file_name );
			watch = itf->m_watch;
		}
		else
		{
			cout<<"sersync插件设置错误，请选择正确的插件名称"<<endl;
			exit(1);
		}

	}

	boost::thread_group syncThreads;
	for (int i = 1; i <= init->sync_num; i++)
	{
		syncThreads.create_thread( boost::bind( &ServerSyncThread, qf, qr, itf ) );
	}
	boost::thread retryThread( boost::bind( &QueueRetryThread, qr ) );


}

int FileSynchronize::XmlParse(boost::shared_ptr<Initialize> init)
{
	XMLNode xMainNode = XMLNode::openFileHelper( init->config_file_name.c_str( ), "head" );
	XMLNode xNode = xMainNode.getChildNode( "sersync" );

	string temp = xNode.getChildNode( "crontab" ).getAttribute( "start" );
	if (temp == "true")
	{
		cout << "开启crontab功能" << "\t";
		m_crontab = atoi( xNode.getChildNode( "crontab" ).getAttribute( "schedule" ) );
		if (m_crontab > 0)
		{
			cout << "每隔" << m_crontab << "分钟整体同步一次" << endl;
		}
	}

	temp = xNode.getChildNode( "plugin" ).getAttribute( "start" );
	if (temp == "true")
	{
		cout << "同步过程执行插件" << "\t";
		m_plugin = xNode.getChildNode( "plugin" ).getAttribute( "name" );
		cout << "插件名称为: " << m_plugin << endl;
	}

	XMLNode lNode = xNode.getChildNode( "localpath", 0 );
	int wtchnum = xNode.nChildNode( "localpath" ); //get the numbers of the localpath tag
	if (wtchnum == 0)
	{
		perror( "错误，配置文件中没有指明需要监控的路径\n" );
		exit( 1 );
	}
	watch = xNode.getChildNode( "localpath" ).getAttribute( "watch" );
	watch = Initialize::SplitLastSlash( watch );
	int remotenum = lNode.nChildNode( "remote" );
	if (remotenum == 0)
	{
		perror( "错误，配置文件中没有指明目标服务器\n" );
		exit( 1 );
	}
	for (int i = 0; i < remotenum; i++)
	{
		string temp_ip = lNode.getChildNode( "remote", i ).getAttribute( "ip" );
		string temp_module = lNode.getChildNode( "remote", i ).getAttribute( "name" );
		if (temp_ip == init->hostip) continue; //ignore the local hostip
		if (temp_ip == "" || temp_module == "")
		{
			perror( "配置文件目标服务器IP或模块名为空" );
			exit( 1 );
		}
		rmtServers.push_back( ptrRmtServer( new RemoteServer( temp_ip, temp_module ) ) );
	}

	if (rmtServers.size( ))
	{
		cout << "同步配置文件解析成功" << endl;
		cout << "建议手动修改/etc/rsyncd.conf中的max connections为0,取消最大连接限制" << endl;
		cout << "程序等待时线程数: " << init->sync_num + 2 << "  = " << "1(主线程) + 1(出错重传线程) + " << init->sync_num \
            << "(开启的同步守护线程) " << endl;
//		cout << "同步文件时开启的最大线程数为: " << 2 + init->sync_num + (rmtServers.size( )) * init->sync_num << "  = " << 2 + init->sync_num << "(等待时开启的线程数) + "\
//              << (rmtServers.size( ))*(init->sync_num) << "(子同步线程)" << endl;
		cout << "请根据同步极限情况的CPU使用率，使用-n参数 进行调整" << endl;
		return 1;
	} else
	{
		return 0; //fail
	}
}

void* FileSynchronize::ServerSyncThread(ptrQFilter qf, ptrQRetry qr, ptrInterface itf)
{
	while (1)
	{
		boost::mutex::scoped_lock nlock( worklock );
		sleep_group++;
		while (qf->empty( ))
		{
			work_cond.wait( nlock );
		}
		sleep_group--;
		Event event = qf->pop( );
		nlock.unlock( );
		boost::thread_group rsync_group;
		for (int offset = 0; offset < rmtServers.size( ); offset++)
		{
			rsync_group.create_thread( boost::bind( &RsyncThread, qr, event, rmtServers[offset]->ip, rmtServers[offset]->module, watch ) );

		}
		rsync_group.join_all( );
		if (itf.use_count( ) != 0)
		{
			itf->Execute( event );
		}

	}
}//end ServerSyncThread

void* FileSynchronize::RsyncThread(ptrQRetry qr, Event event, string ip, string module, string watch)
{
	string eventpath = event->path;
	int operation = event->operation;
	int dir = event->dir;
	string path;
	string command = "cd " + watch + " && rsync -rtu -R ";
	if (operation == 1)
	{
		path = "." + eventpath.substr( watch.length( ), (eventpath.length( ) - watch.length( )) );
		command = command + path + " " + ip + "::" + module + " ";
	} else
	{
		path = eventpath.substr( watch.length( ) + 1, (eventpath.length( ) - watch.length( )) );
		path = DelPathCombine( path, dir );
		command += " --delete ./" + path + " " + ip + "::" + module + " ";
	}
	if (!(debug_level & THREAD_DEBUG))
	{
		command += " >/dev/null 2>&1 ";
	}
	int res = system( command.c_str( ) );
	if (debug_level & THREAD_DEBUG)
	{
		cout << "Rsync command:" << command << endl;
		cout << "res: " << res << endl;
		
	}
	if (res && res != 5888)// 5888 is the error no permitted to  set times  but the file has been rsynced
	{
		qr->push( command );
	}

} //end RsyncThread

void FileSynchronize::ThreadAwaken()
{
	while (!m_qf->empty( ))
	{
		boost::mutex::scoped_lock nlock( worklock );
		if (sleep_group > 0)
		{
			work_cond.notify_one( );
		}
		nlock.unlock( );
	}
}

void* FileSynchronize::QueueRetryThread(ptrQRetry qr)
{
	boost::system_time time_start = boost::get_system_time( );
	boost::system_time crontabStart = boost::get_system_time( );
	while (1)
	{
		boost::system_time timeout = boost::get_system_time( ) + boost::posix_time::minutes( 1 );
		string command;
		bool has_command = qr->time_wait_pop( command, timeout );
		if (!has_command)
		{
			boost::system_time time_cur = boost::get_system_time( );
			boost::posix_time::time_duration offset = time_cur - time_start;
			if (offset.total_seconds( ) >= SCRIPT_EXECUTE_INTERVAL)
			{
				ExecuteScript( qr->fileName );
				time_start = boost::get_system_time( );
			}
			if (m_crontab > 0)
			{
				time_cur = boost::get_system_time( );
				offset = time_cur - time_start;
				if (offset.total_seconds( ) >= (m_crontab * 60))
				{
					RsyncOnce( );
					crontabStart = boost::get_system_time( );
				}
			}

		} else
		{
			int res = system( command.c_str( ) );
			if (res != 0 && res != 5888)// 5888 is the error no permitted to  set times  but the file has been rsynced
			{
				qr->ErrorLog( command ); // only one reader so safe
			}
		}

	}//end while
}

void FileSynchronize::ExecuteScript(string file_name)
{
	string command = "./" + file_name;
	if (debug_level & THREAD_DEBUG)
	{
		cout << command << endl;
	}
	system( command.c_str( ) );
	ofstream out;
	out.open( file_name.c_str( ), ofstream::out ); //清空脚本文件
	out.close( );
	command = "chmod 777 " + file_name;
	system( command.c_str( ) );
}

/**
 *@param str 同步相对路径前面没有./,如 t1/t2/t3.php
 * @return 返回符合rsync 规则的路径 --include=t1/ --include=t1/t2 --include=t1/t2/t3.php --exclude=*
 */
std::string FileSynchronize::DelPathCombine(std::string path, bool dir)
{
	string temp = " ";
	int pos = 0;
	while ((pos = path.find( "/", pos )) != string::npos)
	{
		temp += " --include=" + path.substr( 0, pos + 1 );
		pos++;
	}
	temp += " --include=" + path;
	if (dir == true)
	{
		temp += "  --include=" + path + "/* ";
	}
	temp += " --exclude=* ";
	return temp;
}

void FileSynchronize::RsyncOnce()
{
	static int firstflag = 0;
	if (0 == firstflag)
	{
		cout << "现在对目标服务器统一同步一次" << endl;

	}
	for (int i = 0; i < rmtServers.size( ); i++)
	{
		string command = "cd " + watch + " && rsync -rtu -R --delete ./ ";
		command += rmtServers[i]->ip + "::" + rmtServers[i]->module;
		if (0 == firstflag)
		{
			cout << "执行命令: " << command << endl;
		}
		system( command.c_str( ) );
	}
	if (0 == firstflag)
	{
		cout << "统一推送完成" << endl;
		firstflag++;
	}
}
