#include <dirent.h>
#include<stdio.h>
#include <deque>
#include<sys/select.h>
#include<sys/time.h>
#include<string.h>
#include<fcntl.h>
#include "Inotify.h"
#include"main.h"

//最大缓冲数量
#define MAX_BUF_SIZE 5120
bool printEvent(inotify_event *e);
int flag; //store inotify resource descriptor status;
struct timeval tv;
fd_set rfds;

using namespace std;

Inotify::Inotify(string rootPath)
{
	m_fd = inotify_init( );
	addWatch( rootPath );
	flag = fcntl( m_fd, F_GETFL, 0 );
	tv.tv_sec = 0;
	tv.tv_usec = 100000;
	FD_ZERO( &rfds );
	FD_SET( m_fd, &rfds );

}

unsigned int Inotify::getEvents(ptrQFilter q)
{
	int again = ReadFill( q );
	if (again)
	{
		fcntl( m_fd, F_SETFL, flag | O_NONBLOCK );
		FD_ZERO( &rfds );
		FD_SET( m_fd, &rfds );
		int retval = select( m_fd + 1, &rfds, NULL, NULL, &tv );
		if (retval)
		{
			ReadFill( q );
		}
		fcntl( m_fd, F_SETFL, flag );
	}
	return 1;
}

bool Inotify::ReadFill(ptrQFilter q)
{
	int len = 0;
	char buffer[MAX_BUF_SIZE];
	char* offset = buffer;
	len = read( m_fd, buffer, MAX_BUF_SIZE ); //read May lead to this programme blocking
	if (len == -1)
	{
		return 0;
	}
	while (offset - buffer < len)
	{
		inotify_event* event = (struct inotify_event*) offset;
		offset += sizeof ( struct inotify_event) + event->len;
		if (filterEvent( event ))
		{
			continue;
		}
		if (debug_level & INOTIFY_DEBUG)
		{
			printEvent( event );
		}
		fillEvent( event, q );
	}
	return 1;
}

int Inotify::fillEvent(inotify_event* event, ptrQFilter que)
{
	Event tempEvent( new InotifyEvent( ) );
	tempEvent->wd = event->wd;
	tempEvent->path = (m_path[ event->wd ] + "/" + event->name); //sys sometimes produce large ie.wd which we don't add in the wd list
	tempEvent->mask = event->mask;

	if (event->mask & IN_ISDIR)
	{
		tempEvent->dir = true;
	} else
	{
		tempEvent->dir = false;
	}

	if (tempEvent->dir && ((event->mask & IN_MOVED_FROM) || (event->mask & IN_DELETE)))//remove or delete directory
	{
		removeWatch( tempEvent->path );
		tempEvent->operation = 0;
	}		//create or move  a directory to the watch path
	else if (tempEvent->dir && ((event->mask & IN_MOVED_TO) || (event->mask & IN_CREATE)))
	{
		addWatch( tempEvent->path );
		tempEvent->operation = 1;
	} else
	{
		if ((event->mask & IN_CLOSE_WRITE) || (event->mask & IN_MODIFY) ||
				(event->mask & IN_CREATE) || (event->mask & IN_MOVED_TO)
				)
		{
			tempEvent->operation = 1;
		} else if ((event->mask & IN_MOVED_FROM) || (event->mask & IN_DELETE))
		{
			tempEvent->operation = 0;
		} else if ((event->mask & IN_IGNORED))
		{
			tempEvent->operation = -1;
			return -1;
		}
	}
	que->push( tempEvent );
}

bool Inotify::addWatch(std::string path)
{
	int wd = inotify_add_watch( m_fd, path.c_str( ), IN_SYNC );
	if (-1 == wd)
	{
		perror( "inotify_add_watch error" );
	}
	m_path.insert( PathPair( wd, path ) );
	DIR* pdir = NULL;
	if (!(pdir = opendir( path.c_str( ) )))
	{
		return false;
	}
	struct dirent *pfile = NULL;
	while ((pfile = readdir( pdir ))) //读取目录内容，递归添加监控
	{
		if (pfile->d_type == 4 && strcmp( pfile->d_name, "." ) && strcmp( pfile->d_name, ".." ) && (pfile->d_name[ 0 ] != '.'))
		{
			char buf[ 1024 ];
			sprintf( buf, "%s/%s", path.c_str( ), pfile->d_name );
			addWatch( buf );
		}
	}
	closedir( pdir );
	return true;
}

/*
 * @string  不以斜杠结尾的路径如 /opt/tongbu/temp
 *remove the path from watching deque
 * success return true;
 */
bool Inotify::removeWatch(std::string path)
{
	bool finded = false;
	for (PathMap::iterator i = m_path.begin( ); i != m_path.end( ); i++)
	{
		if (i->second.substr( 0, path.length( ) ) == path) //查找已经添加的监控
		{
			if ((i->second[ path.length( ) ]) == '/' || i->second[ path.length( ) ] == 0)//ignore the error /1234/123/  1234/12/ as the same directory tree
			{
				inotify_rm_watch( m_fd, i->first );
				m_path.erase( i );
				finded = true; //找到并删除
			}
		}
	}
	if (!finded)
	{
		return false;
	}
	return true;
}

int Inotify::filterEvent(struct inotify_event* offset)
{
	int isfilter = 0;
	do
	{
		isfilter = (NULL == offset || (offset ->mask & IN_IGNORED));
		if (isfilter)
		{
#ifdef DEBUG
			if (offset != NULL)
				cout << "offset is: " << offset->name << endl;
			else
				cout << "offset is null" << endl;
#endif
			break;
		}
		isfilter = (strcmp( offset->name, "4913" ) == 0) && ((offset->mask & IN_ISDIR) != 0);
		isfilter = isfilter || (strcmp( (offset->name), "" ) == 0);
		if (isfilter)
		{
#ifdef DEBUG
			cout << "the file name is empty or the file name is 4913:" << offset->name << endl;
#endif
			break;
		}

		int length = strlen( (offset) ->name );
		isfilter = (((offset) ->name[ length - 1 ]) == '~') || ((offset) ->name[ 0 ] == '.');
		if (isfilter)
		{
#ifdef DEBUG
			cout << "this is temporary file name is: " << offset->name << endl;
#endif
			break;
		}

		isfilter = (0 == offset->wd || 0 == offset->mask);
		if (isfilter)
		{
#ifdef DEBUG
			cout << "offset wd is 0 or offset mask is 0; name is: " << offset->name << endl;
			cout << "wd is:" << offset->wd << "\t" << "mask is: " << offset->mask << endl;
#endif
			break;
		}

		PathMap::iterator last = m_path.end( );
		isfilter = ((offset->wd) > (last->first));
		if (isfilter)
		{
#ifdef DEBUG
			cout << endl << "wd out of range not in the pathmap: " << offset->wd << endl;
			cout << "the out of range name is " << offset->name << endl;
			cout << "out of range  mask is " << offset->mask << endl << endl;
#endif
			break;
		}

	} while (0);
	return isfilter;
}

bool printEvent(inotify_event *e)
{
	if (debug_level & SUB_CLASS)
	{
		cout << "name: " << e->name << "\t" << "mask: " << e->mask << "\t" << "len: " << e->len << endl;
	}

}








